/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.NoDatabaseAssociativityException;
import org.ucam.srcf.assassins.exceptions.ObjectModifiedException;

/**
 * Class to represent actual, physical players. Information stored here is
 * shared between all roles (player, police, etc) that a person might have
 * at different points in a game.
 * @see Assassin
 *
 * @author Matthew Johnson <autoumpire@matthew.ath.cx>
 */
public class Player extends CanonicalEntity
{
   static final int READALL = 0;
   static final int READDEF = 1;
   static final int READASS = 2;
   
   /** A Map of all currently instantiated player objects */
   private static Map playerCache = new HashMap();
   /** Objects which have been created and not committed */
   private static List newPlayerCache = new Vector();

   /** 
    * Gets a list of all the players from the database 
    *
    * @return A List of Player objects
    * 
    * @throws AssassinSQLException Thrown if there is a database access error, 
    */
   static List getAllPlayers() throws AssassinSQLException
   {
      // get all the pids
      String query = "SELECT pid, name, email, college, address, waterstatus, password, notes FROM " + DatabaseBackend.getPrefix() + "_players";
      ResultSet r = DatabaseBackend.query(query);
      
      List players = new Vector();
      try {
         while (r.next()) {
            int pid = r.getInt("pid");
            Integer PID = new Integer(pid);

            // check the cache, and add it if necessary
            if (playerCache.containsKey(PID))
               players.add(playerCache.get(PID));
            else {
               Player p = new Player(pid);
               p.name = r.getString("name");
               p.college = r.getString("college");
               p.email = r.getString("email");
               p.address = r.getString("address");
               p.waterstatus = r.getString("waterstatus");
               p.password = r.getString("password");
               p.notes = r.getString("notes");
               players.add(p);
               playerCache.put(PID, p);            
            }

         }
         r.close();
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      return players;
   }


   /** 
    * Loads all the assassins in one go
    */
   static void cacheAssassinsForPlayers() throws AssassinSQLException
   {
      String query = "SELECT aid, pid FROM "+DatabaseBackend.getPrefix()+"_assassins";
      try {
         ResultSet r = DatabaseBackend.query(query);
         while (r.next()) {
            Player p = (Player) playerCache.get(new Integer(r.getInt("pid")));
            if (null == p) continue;
            Assassin a = new Assassin(r.getInt("aid"), p);
            if (null == p.assassins) p.assassins = new Vector();
            p.assassins.add(a);
         }
         r.close();
      } catch (SQLException SQLE) { throw new AssassinSQLException(); }
   }
      

   /** 
    * Gets a list of all the players from the database, ordered by name
    *
    * @return A List of Player objects
    * 
    * @throws AssassinSQLException Thrown if there is a database access error, 
    */
   static List getAllPlayersOrderByName() throws AssassinSQLException
   {
      // get all the pids
      String query = "SELECT pid, name, email, college, address, waterstatus, password, notes FROM " + DatabaseBackend.getPrefix() + "_players ORDER BY name";
      ResultSet r = DatabaseBackend.query(query);
      
      List players = new Vector();
      try {
         while (r.next()) {
            int pid = r.getInt("pid");
            Integer PID = new Integer(pid);

            // check the cache, and add it if necessary
            if (playerCache.containsKey(PID))
               players.add(playerCache.get(PID));
            else {
               Player p = new Player(pid);
               p.name = r.getString("name");
               p.college = r.getString("college");
               p.email = r.getString("email");
               p.address = r.getString("address");
               p.waterstatus = r.getString("waterstatus");
               p.password = r.getString("password");
               p.notes = r.getString("notes");
               players.add(p);
               playerCache.put(PID, p);            
            }

         }
         r.close();
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      return players;
   }

   /** 
    * Gets a list of all the players as Viewers
    *
    * @return A list of PlayerViewer objects
    *
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested pid wasn't found.
    */
   public static List getAllPlayerViewers() throws AssassinSQLException
   {
      // get all the pids
      String query = "SELECT pid FROM " + DatabaseBackend.getPrefix() + "_players";
      ResultSet r = DatabaseBackend.query(query);
      List players = new Vector();
      try { 
         while (r.next()) {
            int pid = r.getInt("pid");
            Integer PID = new Integer(pid);

            // check the cache and add as necessary
            if (playerCache.containsKey(PID)) 
               players.add(new PlayerViewer((Player) playerCache.get(PID)));
            else {
               Player p = new Player(pid);
               playerCache.put(PID, p);            
               players.add(new PlayerViewer(p));
            }
         }
         r.close();
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      return players;
   }
      
   /**
    * Creates a player by pid, checking existance in the database
    *
    * @param pid The database key pid to create from
    * @return The Player object corresponding to the given pid
    * 
    * @throws AssassinSQLException Thrown if there is a database access error, 
    * or if the requested pid wasn't found.
    */
   static Player getByPid(int pid) throws AssassinSQLException
   {
      if (0 == pid) return null;
      Player p;

      // check cache
      if (null == (p = (Player) playerCache.get(new Integer(pid)))) {
         // check database existence
         // String query = "SELECT pid FROM " + DatabaseBackend.getPrefix() + "_players WHERE pid = '"+(new Integer(pid)).toString()+"'";
         // ResultSet r = DatabaseBackend.query(query);
         // try {
         //    if (!r.next()) throw new NonExistentEntityException();
         //    r.close();
         // } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }

         // add to cache
         p = new Player(pid);
         playerCache.put(new Integer(pid), p);
      }
      return p;
   }

   /** Called via reflection to get an instance of this by database ID */
   static Object getByID(Integer id) throws AssassinSQLException
   {
      return getByPid(id.intValue());
   }
   
   /**
    * Removes a player from the database and the pending list.
    *
    * Be Careful with this! Do not reuse a reference if you have deleted it!!
    */
   static void deletePlayer(Player p) throws AssassinSQLException
   {
      newPlayerCache.remove(p);
      playerCache.remove(p);
      if (0 != p.id) {
         String query = "DELETE FROM "+DatabaseBackend.getPrefix()+"_players WHERE pid = '"+p.id+"'";
         try{
			 DatabaseBackend.query(query).close();
         } catch (SQLException e) {
			 throw new AssassinSQLException(e);
         }
         p.id = 0;
      }
   }

   /**
    * Checks all Player objects for updates, and writes them out to the database.
    */
   static void commitAll() throws AssassinSQLException
   {
      // iterate over all new Player objects
      while (newPlayerCache.size() > 0) {
         Player a = (Player) newPlayerCache.remove(0);
         a.writeDetails();
         playerCache.put(new Integer(a.getPid()),a);
      }
      
      // iterate over all cached Player objects
      Iterator i = playerCache.values().iterator();
      while (i.hasNext()) {
         Player p = (Player) i.next();
         // if they are modified, then update
         if (p.dirty()) p.writeDetails();
      }
   }

   /** Name */
   private String name = null;
   /** E-mail */
   private String email = null;
   /** College, or other affiliation */
   private String college = null;
   /** Real Address */
   private String address = null;
   /** Room water status */
   private String waterstatus = null;
   /** Other Notes */
   private String notes = null;
   /** CGI password */
   private String password = null;
   /** A list of all assassins associated with this player */
   private List assassins = null;

   /** Create a blank Player class for adding a player */
   public Player() 
   {
      this.id = 0;
      this.name = "";
      this.email = "";
      this.college = "";
      this.address = "";
      this.waterstatus = "";
      this.notes = "";
      this.password = "";
      this.assassins = new Vector();
      clean();
      newPlayerCache.add(this);
   }
   
   /** Create a Player by database ID */
   Player(int pid)
   {
      this.id = pid;
      clean();
   }

   /** Create a Player by database ID */
   Player(Integer pid)
   {
      this(pid.intValue());
   }

   /** returns the database pid */
   int getPid() { return id; }
    
   /**
    * Retrieve the password of this player.
    * @return The player's password.
    */
   public String getPassword() throws AssassinSQLException
   {
      if (null == password) readDetails(READDEF);
      return password;
   }
   /**
    * Set this player's password.
    * @param password The player's new password.
    */
   public void setPassword(String password)
   {
      setDirty();
      this.password = password;
   }

  
   /**
    * Retrieve the real name of this player.
    * @return The player's name.
    */
   public String getName() throws AssassinSQLException
   {
      if (null == name) readDetails(READDEF);
      return name;
   }
   /**
    * Set this player's real name.
    * @param name The player's new real name.
    */
   public void setName(String name)
   {
      setDirty();
      this.name = name;
   }

   /**
    * Retrieve this player's email address.
    * @return The player's email address.
    */
   public String getEmail() throws AssassinSQLException
   {
      if (null == email) readDetails(READDEF);
      return email;
   }
   /**
    * Set the email address stored for this player.
    * @param email The new email address.
    */
   public void setEmail(String email)
   {
      setDirty();
      this.email = email;
   }
   
   /**
    * Retrieve the college of this player. This is a string rather than an index
    * in order to deal with players who might want to be listed as
    * &quot;ex-Caius&quot; or whatever.
    * @return A string naming the player's college.
    */
   public String getCollege() throws AssassinSQLException
   {
      if (null == college) readDetails(READDEF);
      return college;
   }
   /**
    * Set the player's college.
    * @param college The new string to store for the player's college.
    * @see getCollege()
    */
   public void setCollege(String college)
   {
      setDirty();
      this.college = college;
   }

   /**
    * Retrieve the player's address.
    * @return The player's address.
    */
   public String getAddress() throws AssassinSQLException
   {
      if (null == address) readDetails(READDEF);
      return address;
   }
   /**
    * Set the player's address.
    * @param address The new value for the player's address.
    */
   public void setAddress(String address)
   {
      setDirty();
      this.address = address;
   }

   /**
    * Retrieve the player's preferences for what weapons can be used in their
    * room.
    * @return A string describing the water weapons status.
    */
   public String getWaterStatus() throws AssassinSQLException
   {
      if (null == waterstatus) readDetails(READDEF);
      return waterstatus;
   }
   /**
    * Set this player's water weapons status.
    * @param waterstatus The new description of the player's room status.
    */
   public void setWaterStatus(String waterstatus)
   {
      setDirty();
      this.waterstatus = waterstatus;
   }   
   
   /**
    * Retrieve the &quot;random other information&quot; field for this player.
    * @return The player's notes.
    */
   public String getNotes() throws AssassinSQLException
   {
      if (null == notes) readDetails(READDEF);
      return notes;
   }
   /**
    * Set the notes field for this player.
    * @param notes The new notes to store.
    */
   public void setNotes(String notes)
   {
      setDirty();
      this.notes = notes;
   }
   
   /**
    * Remove an assassin from this Player.
    *
    * @param a The assassin to remove
    */
   public void removeAssassin(Assassin a) throws AssassinSQLException
   {
      if (null == assassins) readDetails(READASS);
      setDirty();
      assassins.remove(a);
      a.setPlayer(null);
   }

   /**
    * Add an assassin to this Player.
    *
    * @param a The assassin to add
    */
   public void addAssassin(Assassin a) throws AssassinSQLException
   {
      if (null == assassins) readDetails(READASS);
      setDirty();
      assassins.add(a);
      a.setPlayer(this);
   }

   /**
    * Return a list of all assassins associated with this player. Each assassin
    * represents an incarnation of the player into a game-role (which may span
    * several actual deaths and resurrections of the 'same' assassin). To
    * modify this list you will need to use {@link Assassin#setPlayer(Player p)}
    * on the assassins in question.
    * @return A list of assassin objects associated with this player.
    */
   public List getAssassins() throws AssassinSQLException
   {
      if (null == assassins) readDetails(READASS);
      return assassins;
   }

   
   /**
    * Return a Viewer object for this object
    *
    * @return A corresponding PlayerViewer object
    */
   public Viewer getViewer()
   {
      return new PlayerViewer(this);
   }

   /**
    * Reads all the details in of the player from the database
    */
   void readDetails() throws AssassinSQLException
   {
      readDetails(READALL);
   }
   void readDetails(int type) throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(Player.class, "Reading Details ("+type+")");
      if (dirty()) throw new ObjectModifiedException();

      if (0 == id) throw new NoDatabaseAssociativityException();
      try { 
         // get all the details of this player
         if (READALL == type || READDEF == type) {
            String query = "SELECT name, email, college, address, waterstatus, password, notes FROM " + DatabaseBackend.getPrefix() + "_players WHERE pid = '"+(new Integer(id)).toString()+"'";
            ResultSet r = DatabaseBackend.query(query);
            if (! r.next()) throw new NoDatabaseAssociativityException();
            this.name = r.getString("name");
            this.college = r.getString("college");
            this.email = r.getString("email");
            this.address = r.getString("address");
            this.waterstatus = r.getString("waterstatus");
            this.password = r.getString("password");
            this.notes = r.getString("notes");
			r.close();
         }
         // get all the assassins associated with this player
         if (READALL == type || READASS == type) {
            String query = "SELECT aid FROM " + DatabaseBackend.getPrefix() + "_assassins WHERE pid = '"+id+"'";
            ResultSet r = DatabaseBackend.query(query);
            assassins = new Vector();
            while (r.next()) {
               int aid = r.getInt("aid");
               Assassin a = Assassin.getByAid(aid);
               if (null == a.getPlayer()) {
                  a.setPlayer(this);
                  clean();
               }
               if (Debug.debug) Debug.print(Player.class, "adding assassin "+a);
               assassins.add(a);
            }
            r.close();
         }
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      clean();
   }

   /**
    * Updates the details of the player in the database
    */
   void writeDetails() throws AssassinSQLException
   {
      if (!dirty()) return;
      clean();
      if (newPlayerCache.contains(this)) newPlayerCache.remove(this);

      try {
         // new object
         if (0 == id) {
            // insert object
            String query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_players (name, college, email, address, waterstatus, notes, password) VALUES ('"+Tools.escape(name)+"','"+Tools.escape(college)+"','"+Tools.escape(email)+"','"+Tools.escape(address)+"','"+Tools.escape(waterstatus)+"', '"+Tools.escape(notes)+"','"+Tools.escape(password)+"')";
            ResultSet r = DatabaseBackend.query(query);
            if (Debug.debug) Debug.print(Player.class, "r.close()");
            r.close();
            query = "SELECT  LAST_INSERT_ID() FROM " + DatabaseBackend.getPrefix() + "_players";
            // get auto-assigned pid
            r = DatabaseBackend.query(query);
            if (!r.next()) throw new AssassinSQLException("Oh Ek! That shouldn't happen?!");
            id = r.getInt(1);
            if (Debug.debug) Debug.print(Player.class, "r.close()");
            r.close();
            if (Debug.debug) Debug.print(Player.class, "r.close() exited");

            // insert assassins
            Iterator i = assassins.iterator();
            while (i.hasNext()) {
               Assassin a = (Assassin) i.next();
               a.setDirty();
               a.writeDetails();
               if (Debug.debug) Debug.print(Player.class, "adding assassin "+a);
               /*query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_assassins (aid, pid) VALUES ('"+a.getAid()+"','"+id+",)";
                 DatabaseBackend.query(query);*/
            }

         }

         // update object
         else {
            String query = "UPDATE " + DatabaseBackend.getPrefix() + "_players SET ";
            if (null != this.name) query += "name = '" + Tools.escape(this.name) + "', ";
            if (null != this.college) query += "college = '" + Tools.escape(this.college) + "', ";
            if (null != this.email) query += "email = '" + Tools.escape(this.email) + "', ";
            if (null != this.address) query += "address = '" + Tools.escape(this.address) + "', ";
            if (null != this.waterstatus) query += "waterstatus = '" + Tools.escape(this.waterstatus) + "', ";
            if (null != this.notes) query += "notes = '" + Tools.escape(this.notes) + "', ";
            if (null != this.password) query += "password = '" + Tools.escape(this.password) + "', ";
            query += " pid = '"+id+"' WHERE pid = '"+id+"'";
            DatabaseBackend.query(query).close();

            // insert assassins
            query = "DELETE FROM " + DatabaseBackend.getPrefix() + "_assassins WHERE pid = '"+id+"'";
            DatabaseBackend.query(query).close();
            Iterator i = assassins.iterator();
            while (i.hasNext()) {
               Assassin a = (Assassin) i.next();
               if (Debug.debug) Debug.print(Player.class, "adding assassin "+a);
               a.setDirty();
               a.writeDetails();
               /* query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_assassins (aid, pid) VALUES ('"+a.getAid()+"','"+pid+",)";
                  DatabaseBackend.query(query);*/
            }
         }
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      
      // mark as unmodified
      if (Debug.debug) Debug.print(Player.class, "cleaning()");
      if (Debug.debug) Debug.print(Player.class, "writeDetails() exit");
   }

   public String getPrintable() throws AssassinSQLException
   {
      return getName();
   }

   protected void setDirty()
   {
      try {
         this.readDetails(READALL);
      } catch(AssassinSQLException ASQLe) {}
      super.setDirty();
   }

   /**
    * Bulk load player/assassin connections
    */
   static void associatePlayersWithAssassins() throws AssassinSQLException
   {
      String query = "SELECT aid, pid FROM "+DatabaseBackend.getPrefix()+"_assassins";
      try {
         ResultSet r = DatabaseBackend.query(query);
         while (r.next()) {
            Player p = Player.getByPid(r.getInt("pid"));
            Assassin a = Assassin.getByAid(r.getInt("aid"));
            a.setPlayer(p);
            a.clean();
            if (null == p.assassins) p.assassins = new Vector();
            p.assassins.add(a);
         }
  			r.close();
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
   }
}

