/***************************************************/
/* Class for static Scoring                        */
/* Copyright 2005 Adam Baird Fraser                */
/* Released under the GPL                          */
/*                                                 */
/***************************************************/

package org.ucam.srcf.assassins.strawberry;

import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;

import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

public class AbFStatic implements StaticScoringAlgorithm {

	private static final int KILL_PTS = 10; // Points gained for each kill
	private static final int DEATH_PTS = 10; // Points lost for each death
	private static final int INIT_PTS = 0; // Initial number of points
	private static final int BONUS = 2; // Number of points for each kill made by each victim
	private static final int KILL_DEC = 2; // How much the points for a kill of a player depreciates by
	private static final int KILL_MIN = 2; // Minimum points transfer for a kill

	private static final boolean REVENGE = true; // Whether or not to use revenge rules - http://asbfraser.org.uk/mayweek3.php

	private HashMap assassins;
	private HashMap kills;

	public void init (int n) {
		// This HashMap contains a mapping of AssassinViewers to AssassinProfiles
		this.assassins = new HashMap();
		// This HashMap is a mapping of AssassinViewers to an ArrayList of KillLinks
		this.kills = new HashMap();
	}


	public Map score() {

		Map scores = new HashMap();

		Iterator it = assassins.keySet().iterator();
		while(it.hasNext()) {
			AssassinViewer av = (AssassinViewer) it.next();
			AssassinProfile ap = (AssassinProfile) assassins.get(av);
			int deaths = ap.getDeaths();

			double killPoints = getPoints(av);

			double points = INIT_PTS + killPoints - DEATH_PTS*deaths;

			scores.put(av, new Double(points));
		}

		return scores;

	}

	private double getPoints(AssassinViewer av) {
		
		double killPoints = 0;
		double bonuses = 0;

		// Get list of KillLinks for this player
		ArrayList playerKills = (ArrayList) kills.get(av);
		// If not in kills, then he hasn't killed anybody. No points, then.
		if(playerKills == null) {
			return 0;
		}

		// Iterate through his kills
		Iterator it = playerKills.iterator();
		while(it.hasNext()) {
			KillLink kl = (KillLink) it.next();

			// For each person he's killed, add on the relevent number of kill points
			killPoints = killPoints + getKillPoints(kl.getNum());
			// Add up his bonuses
			double bonusExtra = BONUS*kl.getNum();
			// Revenge rules are being used
			if(REVENGE) {
				// Get list of kills for the target
				ArrayList playerKilled = (ArrayList) kills.get(kl.getTarget());
				// Make sure he's actually made some kills
				if(playerKilled != null) {
					Iterator it2 = playerKilled.iterator();
					while(it2.hasNext()) {
						KillLink kl2 = (KillLink) it2.next();
						// If the target has killed the assassin
						if(kl2.getTarget().equals(av)) {
							// Find the difference in the number of kills of each other
							int diff = kl.getNum() - kl2.getNum();
							// If the target has killed the assassin as many or more times than the assassin has killed the target, the assassin gets no bonus for it
							if(diff <= 0) {
								bonusExtra = 0;
							}
						}
					}
				}
			}

			bonuses = bonuses + bonusExtra;

		}

		return killPoints + bonuses;
	}

	private double getKillPoints(int number) {
		double killPoints = 0;

		for(int i=0; i<number; i++) {
			// Points for a kill decrease by KILL_DEC each time you kill the same person
			int killAdd = KILL_PTS - i*KILL_DEC;

			// A player will always receive at least KILL_MIN points for a kill 
			if(killAdd < KILL_MIN) {
				killPoints = killPoints + KILL_MIN;
			} else {
				killPoints = killPoints + killAdd;
			}
		}

		return killPoints;
	}

	public void addDeath(Death d) throws AssassinSQLException {

		// Get AssassinViewers for the killer and Victim 
		AssassinViewer killer = d.getKiller();
		AssassinViewer victim = d.getVictim();

		// Retrieve the AssassinProfile of the killer 
		AssassinProfile assassin = (AssassinProfile) assassins.get(killer);
		// If this assassin hasn't been seen before, create a new profile for him and add him to assassins
		if(assassin == null) {
			assassin = new AssassinProfile();
			assassins.put(killer, assassin);
		}
		// He's just made a kill - increment his kills!
		assassin.incKills();

		// Retrieve the AssassinProfile of the victim
		AssassinProfile target = (AssassinProfile) assassins.get(victim);
		// If this assassin hasn't been seen before, create a new profile for him and add him to assassins
		if(target == null) {
			target = new AssassinProfile();
			assassins.put(victim, target);
		}
		// He's just died - increment his deaths!
		target.incDeaths();

		// Get an ArrayList of all KillLinks for killer
		ArrayList playerKills = (ArrayList) kills.get(killer);
		// The killer isn't in kills, so this must be his first
		if(playerKills == null) {
			ArrayList killLinks = new ArrayList();
			killLinks.add(new KillLink(d));
			kills.put(killer, killLinks);
			return;
		}

		// Cycle through the death links
		Iterator it = playerKills.iterator();
		while(it.hasNext()) {
			KillLink kl = (KillLink) it.next();
			// If a link between these two is found, increment the number of kills
			if(kl.getTarget().equals(victim)) {
				kl.incNum();
				return;
			}
		}

		// No link has been found, so create a new one
		playerKills.add(new KillLink(d));
	}


	private class AssassinProfile {
		private int kills;
		private int deaths;

		protected AssassinProfile() {
			this.kills = 0;
			this.deaths = 0;
		}

		protected int getKills() {
			return kills;
		}

		protected int getDeaths() {
			return deaths;
		}

		protected void incKills() {
			kills++;
		}

		protected void incDeaths() {
			deaths++;
		}

	}

	private class KillLink {

		private AssassinViewer assassin;
		private AssassinViewer target;
		private int number;

		protected KillLink(Death d) throws AssassinSQLException {
			this.assassin = d.getKiller();
			this.target = d.getVictim();
			// When the link is created one kill has been made
			number = 1;
		}

		protected AssassinViewer getAssassin() {
			return assassin;
		}

		protected AssassinViewer getTarget() {
			return target;
		}

		protected int getNum() {
			return number;
		}

		protected void incNum() {
			number++;
		}
	}

   /**
    * Tells the algorithm about an innocent death(s).
    * It is up to the alogrithm to use it as it sees fit
    * @param who Who killed it/them?
    * @param count How many they killed.
    */
   public void reportInnocents(AssassinViewer who, int count) throws AssassinSQLException {
   }

   /**
    * Tells the algorithm about bonuses.
    * It is up to the algorithm to actually apply them
    * @param who Who gets the bonus.
    * @param value How much is it worth?
    */
   public void giveBonus(AssassinViewer who, double value) throws AssassinSQLException {
   }
}
